@@ -2,4 +2,4 @@ from .file import File
 from .graph import GraphNode, NodeArg, NodeCall, Out, OutDim
 from .node import Node, NodeLet, NodeLetType, NodeType, TypeConverter, With, WithSet
 from .use import Use, UseByLocal, UseByAuthor, UseByWeb
-from .variable import Expr, Operator, OperatorFn, Shape, Variable, is_estimable, is_hint
+from .variable import Expr, Operator, OperatorFn, Shape, Variable, clone_value, is_estimable, is_hint
